import re
import os
from random import choice, random
from Pipeline.pipeline_file_encoder import *
from mido import MidiFile, MidiTrack, tempo2bpm
import time
from config import *
from json import loads, dump



class PipelineGenerator:
    """
    Class for generating AOP to use when rendering musical data trough plugins.
    """

    def __init__(self):
        pass

    def create_midi_pipeline(self, id, src_path, naming_index, midi_tup, plugins, output_folder="."):
        """
        Creates a complete AOP with midi files processed by generators
        """
        songs = self._loop_rec(src_path, naming_index, midi_tup, finding_midi=True)
        pipeline = self._generate_generator_pipeline(id, songs, plugins, org_folder="/".join(list(songs.values())[0][0][0].split("/")[:naming_index+1]))
        with open(f"{output_folder}/{id}.json", "w", encoding="utf-8") as f:
            json_pipeline = loads(pipeline)
            json.dump(json_pipeline, f, ensure_ascii=False, indent=4)

    def _generate_generator_pipeline(self, name, songs, plugins, include_prob=1, org_folder=None):
        """
        Generates pipelines for a collection of songs using the set effectors
        """
        presets = {plugin: [] for plugin, _ in plugins}
        for plugin, preset_folders in plugins:
            for preset_folder in preset_folders:
                presets[plugin] += [f"{preset_folder}/{preset}" for preset in os.listdir(preset_folder)]
        pipelines = []
        for track, info in songs.items():
            stems, bpm = info
            track_data = []
            for stem in stems:
                if random() < include_prob:
                    generator = choice(list(presets.keys()))
                    preset = choice(presets[generator])
                    generator = Generator(generator, preset)
                    # Create Data object
                    data = Data(stem, generator, None)
                    track_data.append(data)
                # Create Pipeline object
            pipeline = Pipeline(track, bpm, track_data)
            pipelines.append(pipeline)
        t = time.localtime()
        pipeline_file = PipelineFile(name, org_folder, time.strftime("%H:%M:%S", t), pipelines)
        json_pipeline = json.dumps(pipeline_file, indent=4, cls=PipelineFileEncoder)
        return json_pipeline

    def create_effector_pipeline(self, id, src_path, naming_index, midi_tup, plugins, output_folder="."):
        """
        Creates AOP with audio and effectors
        """
        songs = self._loop_rec(src_path, naming_index, midi_tup)
        pipeline = self._generate_effector_pipeline(id, songs, plugins, org_folder="/".join(list(songs.values())[0][0][0].split("/")[:naming_index+1]))
        with open(f"{output_folder}/{id}.json", "w", encoding="utf-8") as f:
            json_pipeline = loads(pipeline)
            json.dump(json_pipeline, f, ensure_ascii=False, indent=4)

    def _generate_effector_pipeline(self, name, songs, plugins, include_prob=1, effect_prob=1, org_folder=None):
        """
        Generates a pipeline for a collection of songs using the set effectors
        :param name: The name of the pipeline file to generate
        :param songs: A dictionary of songs_names, list(stem_paths) that will be used in the pipeline
        :param plugins: Tuples with plugins, preset_folders
        :param include_prob: The probability that a stem is included
        :param effect_prob: The probability that effect is applied to a stem
        :return: The pipeline as json
        """
        presets = {plugin: [] for plugin, _ in plugins}
        for plugin, preset_folders in plugins:
            for preset_folder in preset_folders:
                presets[plugin] += [f"{preset_folder}/{preset}" for preset in os.listdir(preset_folder)]
        pipelines = []
        for track, info in songs.items():
            stems, bpm = info
            track_data = []
            for stem in stems:
                # Create data for each stem (if to be included)
                if random() < include_prob:
                    # Create Effector object
                    effect = None
                    if random() < effect_prob:
                        effector = choice(list(presets.keys()))
                        effect_preset = choice(presets[effector])
                        effect = Effect(effector, effect_preset)
                    # Create Data object
                    data = Data(stem, None, [effect])
                    track_data.append(data)
            # Create Pipeline object
            pipeline = Pipeline(track, bpm, track_data)
            pipelines.append(pipeline)
        # Finally, construct a pipeline file and dump to json
        t = time.localtime()
        pipeline_file = PipelineFile(name, org_folder, time.strftime("%H:%M:%S", t), pipelines)
        json_pipeline = json.dumps(pipeline_file, indent=4, cls=PipelineFileEncoder)
        return json_pipeline

    def _loop_rec(self, src_path, name_index, midi_tup, src_root=".", finding_midi=False, _name_key=None, _counter=0):
        """
        Loops the search path recursively and returns a dictionary where a key is any folder
        matching the src_path at the naming_index and the values are all files matching last index in src_path
        :param src_path: Reg-ex search string, directories/files in the search tree are separated by @
        :param name_index: The index in the search string to use as naming keys
        :param midi_tup: A tuple containing of (1): Where to find midi relative to matched files.
        (2): What pattern to use for matching MIDI
        :param src_root: Where to start searching using the src_path
        :param finding_midi: Whether or not the found files are midi
        :param _name_key: To keep track of the naming when iterating recursively - dont supply!
        :param _counter: To keep track of iterations recursively - dont supply!
        :return: None
        """
        _counter = _counter + 1
        return_paths = {}
        split_path = src_path.split("@")
        rec_left = len(split_path)
        if rec_left > 1:
            # Find the next folder and iterate
            path = os.walk(src_root)
            for root, directories, _ in path:
                for directory in directories:
                    if _counter == name_index - 1:
                        _name_key = directory
                    if re.match(split_path[0], directory):
                        new_src_root = f"{root}/{directory}"
                        returned_files = self._loop_rec('@'.join(split_path[1:]), name_index, midi_tup, new_src_root, finding_midi, _name_key, _counter)
                        return_paths.update(returned_files)
            return return_paths
        else:
            # Voila! You are in the .wav file folder
            path = os.walk(src_root)
            bpm = None
            if not finding_midi:
                # Assuming all .wav in this folder follow the same tempo, it suffices to get the tempo of one of the midis
                bpm = self.find_tempo_from_midi_rec(midi_tup, src_root)
            for root, _, files in path:
                return_files = []
                for file in files:
                    if re.match(split_path[0], file):
                        if not bpm:
                            bpm = self.find_tempo_from_midi(f"{root}/{file}")
                        return_files.append(f"{root}/{file}")
                if return_files:
                    return {_name_key: (return_files, bpm)}
                else:
                    return {}

    def find_tempo_from_midi_rec(self, src_tup, src_root="./"):
        """
        Finds the tempo from a midi found at src_tup[0] relative to src_root that matches src_tup[1]
        :param src_tup: Tuple with (1): Where midis are find relative to src_root, (2) Regex for midi
        :param src_root: Where to start searching for midi
        :return: The tempo of the midi files found first
        """
        path = os.walk(f"{src_root}/{src_tup[0]}")
        for root, _, files in path:
            for file in files:
                if re.match(src_tup[1], file):
                    midi = f"{root}/{file}"
                    return self.find_tempo_from_midi(midi)
        raise FileNotFoundError("No midi file found...")

    def find_tempo_from_midi(self, file_path):
        midi = MidiFile(file_path)
        return self._get_tempo_from_midi(midi)

    @staticmethod
    def _get_plugin_preset_dict(plugin_tup):
        presets = {plugin: [] for plugin, _ in plugin_tup}
        for plugin, preset_folders in plugin_tup:
            for preset_folder in preset_folders:
                presets[plugin] += [f"{preset_folder}/{preset}" for preset in os.listdir(preset_folder)]
        return presets

    @staticmethod
    def _get_tempo_from_midi(midi):
        for track in midi.tracks:
            for msg in track:
                if msg.type == "set_tempo":
                    bpm = tempo2bpm(msg.tempo)
                    return round(bpm)
        return


if __name__ == "__main__":
    pg = PipelineGenerator()
    pg.create_effector_pipeline("TestMidi", SLAKH_LOCAL, SLAKH_LOCAL_NAMING_INDEX, SLAKH_MIDI, [RC20])
    #pg.create_effector_pipeline("Test", URMP_LOCAL, URMP_NAMING_INDEX_LOCAL, URMP_MIDI, [RC20])

