import dawdreamer as daw
from scipy.io import wavfile
from mido import MidiFile, tick2second, second2tick
import os
from json import load, dump, loads
from tqdm import tqdm
from utils import _zero_pad_audio, _merge_audio, _load_audio_file, rshift_midi, match_target_amplitude_of_files


class PipelineParser:

    def __init__(self, sample_rate=44100, buffer_size=128, debug=False):
        self._sample_rate = sample_rate
        self._buffer_size = buffer_size
        self._debug = debug
        self.engine = daw.RenderEngine(sample_rate, buffer_size)
        self._processors = {}

    def _process_midi_with_generator(self, generator_path, midi, preset):
        """
        Creates a generator processor for the supplied midi, using the given preset
        :param generator_path: The path of the generator
        :param midi: The path of the midi
        :param preset: The path of the preset
        :return: A PluginProcessor for the generator
        """
        if generator_path not in self._processors:
            generator = self.engine.make_plugin_processor("generator", generator_path)
            self._processors[generator_path] = generator
        else:
            generator = self._processors[generator_path]
        generator.load_preset(preset)
        generator.load_midi(midi)
        return generator

    def _add_playback_to_graph(self, audio):
        """
        Creates a playback processor for the supplied audio.
        :param audio: The audio to process with playback object (filepath)
        :return: The new graph
        """
        playback = self.engine.make_playback_processor("playback", _load_audio_file(audio, self._sample_rate))
        return playback

    def _add_effectors_to_graph(self, graph, effectors, starting_input):
        """
        Takes a list of effector dictionaries (with vst and presets), and chains them together sequentially in
        the graph, starting from the supplied starting input.
        :param graph: List representing DAG, with tuples as nodes
        :param effectors: List of effectors, with paths and presets
        :param starting_input: The name of the processor to send to first effector
        :return: The new graph
        """
        effector_input = starting_input
        for i in range(len(effectors)):
            effector_name = f"effector{i}"
            if effectors[i]["vst"] not in self._processors:
                effector = self.engine.make_plugin_processor(effector_name, effectors[i]["vst"])
                self._processors[effectors[i]["vst"]] = effector
            else:
                effector = self._processors[effectors[i]["vst"]]
            effector.load_preset(effectors[i]["preset"])
            graph.append((effector, [effector_input]))
            effector_input = effector_name
        return graph

    def _render_graph(self, graph, duration):
        """
        Takes a graph object, and then renders audio trough the processors of the graph, with the given duration.
        :param graph: The graph
        :param duration: The duration of rendered audio.
        """
        self.engine.load_graph(graph)
        self.engine.render(duration)
        audio = self.engine.get_audio()
        return audio

    def process_pipelines(self, instructions_path, output_path=".", overwrite_existing=False):
        """
        Processes a pipeline file. For each operation in the pipeline, an audio output is created.
        Files are stored like ./<PipelineId>/<OperationId>.wav
        :param instructions_path: The file-path containing the instructions for pipelines to process
        :param output_path: Root for output folder
        :param overwrite_existing: Whether or not to overwrite existing files in the output path if they are already there
        :return: None
        """
        instructions = self._get_json_from_file(instructions_path)
        output_folder = f"{output_path}/{instructions['id']}/"
        if not os.path.exists(output_folder):
            os.mkdir(output_folder)
        if overwrite_existing:
            existing_audio = []
        else:
            existing_audio = self._get_existing_audio(instructions["pipelines"], output_folder)
        for pipeline in tqdm(instructions["pipelines"]):
            if pipeline["id"] not in existing_audio:
                operation_audio, tag = self._process_pipeline(pipeline)
                wavfile.write("{}/{}.wav".format(output_folder, tag), self._sample_rate, operation_audio.transpose())
                match_target_amplitude_of_files("{}/{}.wav".format(output_folder, tag), f"{instructions['org_folder']}", tag)
            else:
                if self._debug:
                    print(f"Skipping {pipeline['id']}. Already exists.")
        return

    def _process_pipeline(self, pipeline):
        if self._debug:
            print(f"Now processing {pipeline['id']}")
        bpm = pipeline["bpm"]
        self.engine.set_bpm(bpm)
        audio = []
        for operation in pipeline["data"]:
            audio.append(self._process_pipeline_operation(operation))
        # Merge audio from all operations in pipeline
        return_audio = (_merge_audio(audio), pipeline["id"])
        return return_audio

    def _process_pipeline_operation(self, operation):
        graph = []
        file = operation["file"]
        # Apply audio or midi
        if operation["generator"]:
            generator = self._process_midi_with_generator(operation["generator"]["vst"], file,
                                                          operation["generator"]["preset"])
            starting_input = "generator"
            graph.append((generator, []))
            duration = self.calculate_midi_duration(file)
        else:
            playback = self._add_playback_to_graph(file)
            starting_input = "playback"
            graph.append((playback, []))
            duration = self._calculate_audio_duration(file)
        # Add effectors
        if operation["effects"]:
            graph = self._add_effectors_to_graph(graph, operation["effects"], starting_input)
        # Render audio, then trim
        operation_audio = self._render_graph(graph, duration)
        return operation_audio

    def _get_existing_audio(self, pipelines, output_folder):
        existing_audio = []
        for pipeline in pipelines:
            file_path = f"{output_folder}/{pipeline['id']}.wav"
            if os.path.exists(file_path):
                existing_audio.append(pipeline["id"])
        return existing_audio

    def _calculate_audio_duration(self, audio_path):
        """
        Takes an audio file and returns the duration of the longest one in seconds.
        :param audio_path: audio file
        :return: The duration of the audio file in seconds
        """
        audio = _load_audio_file(audio_path, self._sample_rate)
        length = audio.shape[1] / self._sample_rate
        return length

    def calculate_midi_duration(self, midi):
        """
        Calculates the length of a midi file in seconds.
        :param midi: The midi file
        :return: The duration of the midi in seconds.
        """
        midi = MidiFile(midi)
        return midi.length

    @staticmethod
    def _get_json_from_file(file_path, enc="utf-8"):
        """
        Opens a file and returns the data as a json object
        :param file_path: The file path for the json
        :param enc: The encoding to use when reading
        :return: A dictionary representing the json
        """
        with open(file_path, 'r+', encoding=enc,) as file:
            data = load(file)
        return data



if __name__ == "__main__":
    pp = PipelineParser()
    # pp.process_pipelines_over_ssh(SSH_HOST, SSH_USER, SSH_PASSWORD, "./AMT/data.json", "./Datasets")
    #pp.process_pipelines("./data.json")
    pp.process_pipelines("test.json", overwrite_existing=True)




