from paramiko import SSHClient, AutoAddPolicy
import os
import re
from config import *
from pipeline_generator import PipelineGenerator
from mido import MidiFile, tempo2bpm
import sys

"""
A PipelineGenerator that gets datasets from a remote server using SSH. 
The local presets and VSTs are then used in the pipeline for later processing using a SSHPipelineParser
"""


class SSHPipelineGenerator(PipelineGenerator):

    def __init__(self, remote_server, user, password):
        super().__init__()
        self.remote_server = remote_server
        self.ssh_client = self._init_ssh_client(remote_server, user, password)
        self.sftp_client = self.ssh_client.open_sftp()

    @staticmethod
    def _init_ssh_client(remote, user, pwd):
        ssh_client = SSHClient()
        ssh_client.set_missing_host_key_policy(AutoAddPolicy())
        ssh_client.connect(hostname=remote, username=user, password=pwd)
        return ssh_client

    def _loop_rec(self, src_path, name_index, midi_tup, src_root=".", finding_midi=False, _name_key=None, _counter=0):
        _counter += 1
        return_paths = {}
        split_path = src_path.split("@")
        rec_left = len(split_path)
        if rec_left > 1:
            for directory in self.sftp_client.listdir(src_root):
                if _counter == name_index + 1:
                    _name_key = directory
                if re.match(split_path[0], directory):
                    new_src_root = f"{src_root}/{directory}"
                    returned_files = self._loop_rec('@'.join(split_path[1:]), name_index, midi_tup, new_src_root, finding_midi, _name_key,
                                              _counter)
                    return_paths.update(returned_files)
            return return_paths
        else:
            # Voila! You are in the .wav file folder
            # Assuming all .wav in this folder follow the same tempo, it suffices to get the tempo of one of the midis
            try:
                bpm = self.find_tempo_from_midi(midi_tup, src_root)
            except (EOFError, FileNotFoundError) as e:
                print(e)
                print("Defaulting to tempo 120")
                bpm = 120
            return_files = []
            for file in self.sftp_client.listdir(src_root):
                if re.match(split_path[0], file):
                    return_files.append(f"{src_root}/{file}")
            if return_files:
                return {_name_key: (return_files, bpm)}
            else:
                return {}

    def find_tempo_from_midi(self, src_tup, src_root="./"):
        """
        Finds the tempo from a midi found at src_tup[0] relative to src_root that matches src_tup[1]
        :param src_tup: Tuple with (1): Where midis are find relative to src_root, (2) Regex for midi
        :param src_root: Where to start searching for midi
        :return: The tempo of the midi files found first
        """
        for file in self.sftp_client.listdir(f"{src_root}/{src_tup[0]}"):
            if re.match(src_tup[1], file):
                # Need to download file locally, then parse to midi, and find tempo.
                # Before returning, get rid of local midi
                b_midi = self.sftp_client.open(f"{src_root}/{src_tup[0]}/{file}")
                b_midi.prefetch()
                b_midi = b_midi.read()
                with open("temporary.mid", "wb") as f:
                    # b_midi = bytearray([b if b <= 127 else 127 for b in b_midi])
                    f.write(b_midi)
                try:
                    midi = MidiFile("temporary.mid", clip=True)
                    for track in midi.tracks:
                        for msg in track:
                            if msg.type == "set_tempo":
                                bpm = tempo2bpm(msg.tempo)
                                os.remove("temporary.mid")
                                return round(bpm)
                except (EOFError, OSError) as e:
                    exception_str = 'Unexpected error in ' + file + ':\n', e, sys.exc_info()[0]
                    raise EOFError(exception_str)
        raise FileNotFoundError("No midi file found...")


if __name__ == "__main__":
    spp = SSHPipelineGenerator(SSH_HOST, SSH_USER, SSH_PASSWORD)
    #spp.create_effector_pipeline("GuitarSet-SSH", GUITARSET_AUDIO, GUITARSET_NAMING_INDEX, GUITARSET_MIDI_ALL, [RC20])
    #spp.create_effector_pipeline("URMP-RC20", URMP, URMP_NAMING_INDEX, URMP_MIDI_ALL, [RC20])
    #spp.create_effector_pipeline("Musicnet-RC20", MUSICNET_AUDIO, MUSICNET_NAMING_INDEX, MUSICNET_MIDI_ALL, [RC20])
    #spp.create_effector_pipeline("Maestro-RC20", MAESTRO_AUDIO, MAESTRO_NAMING_INDEX, MAESTRO_MIDI_ALL, [RC20])
    #spp.create_effector_pipeline("Cerberus-RC20", CERBERUS_AUDIO, CERBERUS_NAMING_INDEX, CERBERUS_MIDI_ALL, [RC20])
    #spp.create_effector_pipeline("Slakh-RC20", SLAKH_AUDIO, SLAKH_NAMING_INDEX, SLAKH_MIDI_ALL, [RC20])
    #spp.create_effector_pipeline("GuitarSet-SerumFX", GUITARSET_AUDIO, GUITARSET_NAMING_INDEX, GUITARSET_MIDI_ALL, [SERUM_FX])
    #spp.create_effector_pipeline("URMP-SerumFX", URMP, URMP_NAMING_INDEX, URMP_MIDI_ALL, [SERUM_FX])
    #spp.create_effector_pipeline("Musicnet-SerumFX", MUSICNET_AUDIO, MUSICNET_NAMING_INDEX, MUSICNET_MIDI_ALL, [SERUM_FX])
    #spp.create_effector_pipeline("Maestro-SerumFX", MAESTRO_AUDIO, MAESTRO_NAMING_INDEX, MAESTRO_MIDI_ALL, [SERUM_FX])


    #spp.create_midi_pipeline("GuitarSet-Serum", GUITARSET_MIDI, GUITARSET_NAMING_INDEX, GUITARSET_MIDI_ALL, [SERUM])
    #spp.create_midi_pipeline("URMP-Serum", URMP_MIDI, URMP_NAMING_INDEX, URMP_MIDI_ALL, [SERUM])
    spp.create_midi_pipeline("MusicNet-Serum", MUSICNET_MIDI, MUSICNET_NAMING_INDEX, MUSICNET_MIDI_ALL, [SERUM])
    #spp.create_midi_pipeline("Maestro-Serum", MAESTRO_MIDI, MAESTRO_NAMING_INDEX, MAESTRO_MIDI_ALL, [SERUM])
    #spp.create_midi_pipeline("GuitarSet-Kontakt", GUITARSET_MIDI, GUITARSET_NAMING_INDEX, GUITARSET_MIDI_ALL, [KONTAKT])
    #spp.create_midi_pipeline("URMP-Kontakt", URMP_MIDI, URMP_NAMING_INDEX, URMP_MIDI_ALL, [KONTAKT])
    #spp.create_midi_pipeline("MusicNet-Kontakt", MUSICNET_MIDI, MUSICNET_NAMING_INDEX, MUSICNET_MIDI_ALL, [KONTAKT])
    #spp.create_midi_pipeline("Maestro-Kontakt", MAESTRO_MIDI, MAESTRO_NAMING_INDEX, MAESTRO_MIDI_ALL, [KONTAKT])

    #spp.create_effector_pipeline("GuitarSet-RC20-13P", GUITARSET_AUDIO, GUITARSET_NAMING_INDEX, GUITARSET_MIDI_ALL, [RC20_13P])
    #spp.create_effector_pipeline("GuitarSet-RC20-25P", GUITARSET_AUDIO, GUITARSET_NAMING_INDEX, GUITARSET_MIDI_ALL, [RC20_25P])
    #spp.create_effector_pipeline("GuitarSet-RC20-50P", GUITARSET_AUDIO, GUITARSET_NAMING_INDEX, GUITARSET_MIDI_ALL, [RC20_50P])
    #spp.create_effector_pipeline("GuitarSet-RC20-75P", GUITARSET_AUDIO, GUITARSET_NAMING_INDEX, GUITARSET_MIDI_ALL, [RC20_75P])
