from pipeline_parser import PipelineParser
from paramiko import SSHClient, AutoAddPolicy
from json import loads
from config import *
from scipy.io import wavfile
import os
from mido import MidiFile
from utils import _load_audio_file
from tqdm import tqdm
from time import time
from utils import sftp_match_target_amplitude_of_files, match_target_amplitude, _load_audio_file
import re
from pydub import AudioSegment
import numpy as np



class SSHPipelineParser(PipelineParser):

    def __init__(self, remote_server, user, password, temporary_folder="./SSHParserTemp", sample_rate=16000, buffer_size=128, debug=False):
        super().__init__(sample_rate, buffer_size, debug)
        self.remote_server = remote_server
        self._user = user
        self._password = password
        self._temporary_folder = temporary_folder
        self.ssh_client = self._init_ssh_client(remote_server, user, password)
        self.sftp_client = self.ssh_client.open_sftp()

    @staticmethod
    def _init_ssh_client(remote, user, pwd):
        ssh_client = SSHClient()
        ssh_client.set_missing_host_key_policy(AutoAddPolicy())
        ssh_client.connect(hostname=remote, username=user, password=pwd)
        return ssh_client

    def process_pipelines_over_ssh(self,  instructions_path, output_path=".", overwrite_existing=False):
        # Get pipeline-instructions from server
        """
        b_instructions = self.sftp_client.open(instructions_path)
        b_instructions.prefetch()
        b_instructions = b_instructions.read()
        instructions = b_instructions.decode("utf-8")
        instructions = loads(instructions)
        """
        instructions = self._get_json_from_file(instructions_path)
        # Make sure output folder exists
        output_folder = f"{output_path}/{instructions['id']}/"
        if f"{instructions['id']}" not in self.sftp_client.listdir(output_path):
            self.sftp_client.mkdir(output_folder)
        # Create temp folder, then delete after
        if not os.path.exists(self._temporary_folder):
            os.mkdir(self._temporary_folder)
        if overwrite_existing:
            existing_audio = []
        else:
            existing_audio = self._get_existing_audio(instructions["pipelines"], output_folder)
        # Process each operation
        for pipeline in tqdm(instructions["pipelines"]):
            # Don't process a file if its id is supplied as existing audio
            if pipeline["id"] not in existing_audio:
                # Get audio from processing
                pipeline_audio, id = self._process_pipeline(pipeline, instructions["stem_folder"])
                pipeline_audio = pipeline_audio.mean(axis=0) / len(pipeline_audio)
                # Save the file locally and then transfer to remote server
                local_file = f"{self._temporary_folder}/{id}.wav"
                wavfile.write(local_file, self._sample_rate, pipeline_audio.transpose())
                os.system(f"sox {local_file} -b 32 temporary_sox.wav")
                os.remove(local_file)
                os.rename("temporary_sox.wav", local_file)
                sftp_match_target_amplitude_of_files(local_file, f"{instructions['org_folder']}/{id}", id, self.sftp_client)
                self.sftp_client.put(local_file, f"{output_folder}/{id}.wav")
            else:
                if self._debug:
                    print(f"Skipping {pipeline['id']}. Already exists.")
        #os.remove(self._temporary_folder)
        return

    def _normalize_operation_audio(self, audio, file, stem_dir):
        # Get AudioSegment from audio and read volume
        wavfile.write("temporary_normalization_sound.wav", self._sample_rate, audio.transpose())
        sound = AudioSegment.from_file("temporary_normalization_sound.wav")
        # Get the original file from server
        track_folder = "/".join(file.split("/")[:-1])
        file_tag = file.split("/")[-2]
        full_stem_dir = f"{track_folder}/{stem_dir}"
        # norm_audio = sftp_match_target_amplitude_of_files("temporary_normalization_sound.wav", full_stem_dir, file_tag, self.sftp_client)
        for f in self.sftp_client.listdir(full_stem_dir):
            if re.match(f"(.*){file_tag}(.*).wav", f):
                org_track = f"{full_stem_dir}/{f}"
        if not org_track:
            for f in self.sftp_client.listdir(full_stem_dir):
                if re.match("(.*).wav", f):
                    if org_track:
                        raise EnvironmentError("Unexpected number of files to match..")
                    org_track = f"{full_stem_dir}/{f}"
        # Save the file temporarily
        b_org = self.sftp_client.open(org_track)
        b_org.prefetch()
        b_org = b_org.read()
        with open("temporary_normalization_org.wav", "wb") as f:
            f.write(b_org)
        original = AudioSegment.from_file("temporary_normalization_org.wav")
        # Get the volume and normalize
        normalized_sound = match_target_amplitude(sound, original.dBFS)
        normalized_sound.export("temporary_normalized_audio.wav", format="wav")
        return_audio = _load_audio_file("temporary_normalized_audio.wav", SAMPLE_RATE)

        return return_audio


    def _get_existing_audio(self, pipelines, output_folder):
        existing_audio = []
        for pipeline in tqdm(pipelines):
            file_path = f"{output_folder}/{pipeline['id']}.wav"
            try:
                self.sftp_client.lstat(file_path)
                existing_audio.append(pipeline["id"])
            except FileNotFoundError:
                pass
        return existing_audio

    def _process_midi_with_generator(self, generator, midi, preset):
        """
        Overwrites the function from parent by introducing loading of MIDI over SSH.
        Creates a generator processor for the supplied midi, using the given preset
        :param generator: The path of the generator - on local
        :param midi: The path of the midi - on the remote server
        :param preset: The path of the preset - on local
        :return: A PluginProcessor for the generator
        """
        generator = self.engine.make_plugin_processor("generator", generator)
        generator.load_preset(preset)
        b_midi = self.sftp_client.open(midi)
        b_midi.prefetch()
        b_midi = b_midi.read()
        temp_midi = f"{self._temporary_folder}/temporary.mid"
        with open(temp_midi, "wb") as file:
            file.write(b_midi)
        generator.load_midi(temp_midi)
        return generator

    def calculate_midi_duration(self, midi):
        midi = f"{self._temporary_folder}/temporary.mid"
        midi = MidiFile(midi)
        return midi.length

    def _add_playback_to_graph(self, audio):
        """
        Creates a playback processor for the supplied audio.
        :param audio: The audio to process with playback object (remote filepath)
        :return: The new graph
        """
        tempo_audio = f"{self._temporary_folder}/temporary.wav"
        self.sftp_client.get(audio, tempo_audio)
        playback = self.engine.make_playback_processor("playback", _load_audio_file(tempo_audio, self._sample_rate))
        return playback

    def _calculate_audio_duration(self, audio_path):
        """
        Takes an audio file and returns the duration of the longest one in seconds.
        :param audio_path: audio file
        :return: The duration of the audio file in seconds
        """
        audio = f"{self._temporary_folder}/temporary.wav"
        audio = _load_audio_file(audio, self._sample_rate)
        length = audio.shape[1] / self._sample_rate
        return length


if __name__ == "__main__":
    a = SSHPipelineParser(SSH_HOST, SSH_USER, SSH_PASSWORD, sample_rate=SAMPLE_RATE, debug=True)
    #a.process_pipelines_over_ssh("./Guitarset-SSH.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./MusicNet-RC20.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./URMP-RC20.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./Maestro-RC20.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./Cerberus-RC20.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./Slakh-RC20.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./GuitarSet-SerumFX.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./MusicNet-SerumFX.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./URMP-SerumFX.json", output_path="RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./Maestro-SerumFX.json", output_path="RenderedData/Exp1", overwrite_existing=True)

    #a.process_pipelines_over_ssh("./Guitarset-Omnisphere.json", "RenderedData/Exp2", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./Guitarset-Omnisphere.json", "RenderedData/Exp2", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./GuitarSet-Serum.json", "RenderedData/Exp2", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./Maestro-Serum.json", "RenderedData/Exp2")
    #a.process_pipelines_over_ssh("./URMP-Serum.json", "RenderedData/Exp2", overwrite_existing=True)
    a.process_pipelines_over_ssh("./MusicNet-Serum.json", "RenderedData/Exp2", overwrite_existing=True)
    # bga.process_pipelines_over_ssh("./GuitarSet-Kontakt.json", "RenderedData/Exp2", overwrite_existing=True)

    #a.process_pipelines_over_ssh("./URMP-Kontakt.json", "RenderedData/Exp2", overwrite_existing=Truekjhg)
    #a.process_pipelines_over_ssh("./urmp-test.json", "RenderedData/Exp2", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./MusicNet-Kontakt.json", "RenderedData/Exp2", overwrite_existing=False)

    #a.process_pipelines_over_ssh("./GuitarSet-RC20-13P.json", "RenderedData/Exp1")
    #a.process_pipelines_over_ssh("./GuitarSet-RC20-25P.json", "RenderedData/Exp1", overwrite_existing=True)
    #a.process_pipelines_over_ssh("./GuitarSet-RC20-50P.json")
    #a.process_pipelines_over_ssh("./GuitarSet-RC20-75P.json")
    #a.process_pipelines_over_ssh("./Maestro-Kontakt.json", "RenderedData/Exp2", overwrite_existing=False)